#!/usr/bin/env python3

"""
Dumps attention weights.

E.g. to Numpy npy files.
To load them::

    d = np.load("....npy").item()
    d = [v for (k, v) in d.items()]
    att_weights = d[-1]['rec_att_weights'].squeeze(axis=2)
    import matplotlib.pyplot as plt
    plt.matshow(att_weights)
    plt.show()

Or directly as png images.

Or into HDF. In this case, this tool is very similar to `rnn.py --task=forward`.

"""

from __future__ import print_function

import os
import sys
import numpy as np
import argparse
from glob import glob
import typing

# Returnn imports
import _setup_returnn_env  # noqa
import returnn.__main__ as rnn
from returnn.tf.engine import Runner
from returnn.datasets import init_dataset
from returnn.util.basic import NumbersDict, Stats, deep_update_dict_values


def inject_retrieval_code(net_dict, rec_layer_name, layers, dropout):
  """
  Injects some retrieval code into the config

  :param dict[str] net_dict:
  :param str rec_layer_name: name of rec layer
  :param list[str] layers: layers in rec layer to extract
  :param float|None dropout: to override, if given
  :return: net_dict
  :rtype: dict[str]
  """
  assert config is not None
  assert rec_layer_name in net_dict
  assert net_dict[rec_layer_name]["class"] == "rec"
  for layer in layers:
    assert layer in net_dict[rec_layer_name]['unit'], "layer %r not found" % layer

  new_layers_descr = net_dict.copy()  # actually better would be deepcopy...
  for sub_layer in layers:
    # assert that sub_layer inside subnet is a output-layer
    new_layers_descr[rec_layer_name]['unit'][sub_layer]["is_output_layer"] = True

  if dropout is not None:
    deep_update_dict_values(net_dict, "dropout", dropout)
    deep_update_dict_values(net_dict, "rec_weight_dropout", dropout)
  return new_layers_descr


config = None  # type: typing.Optional["returnn.config.Config"]


def init_returnn(config_fn, args):
  """
  :param str config_fn:
  :param args: arg_parse object
  """
  rnn.init_better_exchook()
  config_updates = {
    "log": [],
    "task": "eval",
    "need_data": False,  # we will load it explicitly
    "device": args.device if args.device else None}
  if args.epoch:
    config_updates["load_epoch"] = args.epoch
  if args.do_search:
    config_updates.update({
      "task": "search",
      "search_do_eval": False,
      "beam_size": args.beam_size,
      "max_seq_length": 0,
      })

  rnn.init(
    config_filename=config_fn,
    config_updates=config_updates, extra_greeting="RETURNN get-attention-weights starting up.")
  global config
  config = rnn.config


def init_net(args, layers):
  """
  :param args:
  :param list[str] layers:
  """
  def net_dict_post_proc(net_dict):
    """
    :param dict[str,dict[str]] net_dict:
    :rtype: dict[str,dict[str]]
    """
    return inject_retrieval_code(net_dict, rec_layer_name=args.rec_layer, layers=layers, dropout=args.dropout)

  rnn.engine.use_dynamic_train_flag = True  # will be set via Runner. maybe enabled if we want dropout
  rnn.engine.init_network_from_config(config=config, net_dict_post_proc=net_dict_post_proc)


def main(argv):
  """
  Main entry.
  """
  argparser = argparse.ArgumentParser(description=__doc__)
  argparser.add_argument("config_file", type=str, help="RETURNN config, or model-dir")
  argparser.add_argument("--epoch", type=int)
  argparser.add_argument('--data', default="train",
                         help="e.g. 'train', 'config:train', or sth like 'config:get_dataset('dev')'")
  argparser.add_argument('--do_search', default=False, action='store_true')
  argparser.add_argument('--beam_size', default=12, type=int)
  argparser.add_argument('--dump_dir', help="for npy or png")
  argparser.add_argument("--output_file", help="hdf")
  argparser.add_argument("--device", help="gpu or cpu (default: automatic)")
  argparser.add_argument("--layers", default=["att_weights"], action="append",
                         help="Layer of subnet to grab")
  argparser.add_argument("--rec_layer", default="output", help="Subnet layer to grab from; decoder")
  argparser.add_argument("--enc_layer", default="encoder")
  argparser.add_argument("--batch_size", type=int, default=5000)
  argparser.add_argument("--seq_list", default=[], action="append", help="predefined list of seqs")
  argparser.add_argument("--min_seq_len", default="0", help="can also be dict")
  argparser.add_argument("--num_seqs", default=-1, type=int, help="stop after this many seqs")
  argparser.add_argument("--output_format", default="npy", help="npy, png or hdf")
  argparser.add_argument("--dropout", default=None, type=float, help="if set, overwrites all dropout values")
  argparser.add_argument("--train_flag", action="store_true")
  argparser.add_argument("--reset_partition_epoch", type=int, default=1)
  argparser.add_argument("--reset_seq_ordering", default="sorted_reverse")
  argparser.add_argument("--reset_epoch_wise_filter", default=None)
  args = argparser.parse_args(argv[1:])

  layers = args.layers
  assert isinstance(layers, list)
  config_fn = args.config_file
  explicit_model_dir = None
  if os.path.isdir(config_fn):
    # Assume we gave a model dir.
    explicit_model_dir = config_fn
    train_log_dir_config_pattern = "%s/train-*/*.config" % config_fn
    train_log_dir_configs = sorted(glob(train_log_dir_config_pattern))
    assert train_log_dir_configs
    config_fn = train_log_dir_configs[-1]
    print("Using this config via model dir:", config_fn)
  else:
    assert os.path.isfile(config_fn)
  model_name = ".".join(config_fn.split("/")[-1].split(".")[:-1])

  init_returnn(config_fn=config_fn, args=args)
  if explicit_model_dir:
    config.set("model", "%s/%s" % (explicit_model_dir, os.path.basename(config.value('model', ''))))
  print("Model file prefix:", config.value('model', ''))

  if args.do_search:
    raise NotImplementedError
  min_seq_length = NumbersDict(eval(args.min_seq_len))

  assert args.output_format in ["npy", "png", "hdf"]
  if args.output_format in ["npy", "png"]:
    assert args.dump_dir
    if not os.path.exists(args.dump_dir):
      os.makedirs(args.dump_dir)
  plt = ticker = None
  if args.output_format == "png":
    import matplotlib.pyplot as plt  # need to import early? https://stackoverflow.com/a/45582103/133374
    import matplotlib.ticker as ticker

  dataset_str = args.data
  if dataset_str in ["train", "dev", "eval"]:
    dataset_str = "config:%s" % dataset_str
  extra_dataset_kwargs = {}
  if args.reset_partition_epoch:
    print("NOTE: We are resetting partition epoch to %i." % (args.reset_partition_epoch,))
    extra_dataset_kwargs["partition_epoch"] = args.reset_partition_epoch
  if args.reset_seq_ordering:
    print("NOTE: We will use %r seq ordering." % (args.reset_seq_ordering,))
    extra_dataset_kwargs["seq_ordering"] = args.reset_seq_ordering
  if args.reset_epoch_wise_filter:
    extra_dataset_kwargs["epoch_wise_filter"] = eval(args.reset_epoch_wise_filter)
  dataset = init_dataset(dataset_str, extra_kwargs=extra_dataset_kwargs)
  if hasattr(dataset, "epoch_wise_filter") and args.reset_epoch_wise_filter is None:
    if dataset.epoch_wise_filter:
      print("NOTE: Resetting epoch_wise_filter to None.")
      dataset.epoch_wise_filter = None
  if args.reset_partition_epoch:
    assert dataset.partition_epoch == args.reset_partition_epoch
  if args.reset_seq_ordering:
    assert dataset.seq_ordering == args.reset_seq_ordering

  init_net(args, layers)
  network = rnn.engine.network

  hdf_writer = None
  if args.output_format == "hdf":
    assert args.output_file
    assert len(layers) == 1
    sub_layer = network.get_layer("%s/%s" % (args.rec_layer, layers[0]))
    from returnn.datasets.hdf import SimpleHDFWriter
    hdf_writer = SimpleHDFWriter(filename=args.output_file, dim=sub_layer.output.dim, ndim=sub_layer.output.ndim)

  extra_fetches = {
    "output": network.layers[args.rec_layer].output.get_placeholder_as_batch_major(),
    "output_len": network.layers[args.rec_layer].output.get_sequence_lengths(),  # decoder length
    "encoder_len": network.layers[args.enc_layer].output.get_sequence_lengths(),  # encoder length
    "seq_idx": network.get_extern_data("seq_idx"),
    "seq_tag": network.get_extern_data("seq_tag"),
    "target_data": network.get_extern_data(network.extern_data.default_input),
    "target_classes": network.get_extern_data(network.extern_data.default_target),
  }
  for layer in layers:
    sub_layer = rnn.engine.network.get_layer("%s/%s" % (args.rec_layer, layer))
    extra_fetches["rec_%s" % layer] = sub_layer.output.get_placeholder_as_batch_major()
  dataset.init_seq_order(epoch=1, seq_list=args.seq_list or None)  # use always epoch 1, such that we have same seqs
  dataset_batch = dataset.generate_batches(
    recurrent_net=network.recurrent,
    batch_size=args.batch_size,
    max_seqs=rnn.engine.max_seqs,
    max_seq_length=sys.maxsize,
    min_seq_length=min_seq_length,
    max_total_num_seqs=args.num_seqs,
    used_data_keys=network.used_data_keys)

  stats = {layer: Stats() for layer in layers}

  # (**dict[str,numpy.ndarray|str|list[numpy.ndarray|str])->None
  def fetch_callback(seq_idx, seq_tag, target_data, target_classes, output, output_len, encoder_len, **kwargs):
    """
    :param list[int] seq_idx: len is n_batch
    :param list[str] seq_tag: len is n_batch
    :param numpy.ndarray target_data: extern data default input (e.g. "data"), shape e.g. (B,enc-T,...)
    :param numpy.ndarray target_classes: extern data default target (e.g. "classes"), shape e.g. (B,dec-T,...)
    :param numpy.ndarray output: rec layer output, shape e.g. (B,dec-T,...)
    :param numpy.ndarray output_len: rec layer seq len, i.e. decoder length, shape (B,)
    :param numpy.ndarray encoder_len: encoder seq len, shape (B,)
    :param kwargs: contains "rec_%s" % l for l in layers, the sub layers (e.g att weights) we are interested in
    """
    n_batch = len(seq_idx)
    for i in range(n_batch):
      # noinspection PyShadowingNames
      for layer in layers:
        att_weights = kwargs["rec_%s" % layer][i]
        stats[layer].collect(att_weights.flatten())
    if args.output_format == "npy":
      data = {}
      for i in range(n_batch):
        data[i] = {
          'tag': seq_tag[i],
          'data': target_data[i],
          'classes': target_classes[i],
          'output': output[i],
          'output_len': output_len[i],
          'encoder_len': encoder_len[i],
        }
        # noinspection PyShadowingNames
        for layer in [("rec_%s" % layer) for layer in layers]:
          assert layer in kwargs
          out = kwargs[layer][i]
          assert out.ndim >= 2
          assert out.shape[0] >= output_len[i] and out.shape[1] >= encoder_len[i]
          data[i][layer] = out[:output_len[i], :encoder_len[i]]
        fname = args.dump_dir + '/%s_ep%03d_data_%i_%i.npy' % (model_name, rnn.engine.epoch, seq_idx[0], seq_idx[-1])
        np.save(fname, data)
    elif args.output_format == "png":
      for i in range(n_batch):
        # noinspection PyShadowingNames
        for layer in layers:
          extra_postfix = ""
          if args.dropout is not None:
            extra_postfix += "_dropout%.2f" % args.dropout
          elif args.train_flag:
            extra_postfix += "_train"
          fname = args.dump_dir + '/%s_ep%03d_plt_%05i_%s%s.png' % (
            model_name, rnn.engine.epoch, seq_idx[i], layer, extra_postfix)
          att_weights = kwargs["rec_%s" % layer][i]
          att_weights = att_weights.squeeze(axis=2)  # (out,enc)
          assert att_weights.shape[0] >= output_len[i] and att_weights.shape[1] >= encoder_len[i]
          att_weights = att_weights[:output_len[i], :encoder_len[i]]
          print("Seq %i, %s: Dump att weights with shape %r to: %s" % (
            seq_idx[i], seq_tag[i], att_weights.shape, fname))
          plt.matshow(att_weights)
          title = seq_tag[i]
          if dataset.can_serialize_data(network.extern_data.default_target):
            title += "\n" + dataset.serialize_data(
              network.extern_data.default_target, target_classes[i][:output_len[i]])
            ax = plt.gca()
            tick_labels = [
              dataset.serialize_data(network.extern_data.default_target, np.array([x], dtype=target_classes[i].dtype))
              for x in target_classes[i][:output_len[i]]]
            ax.set_yticklabels([''] + tick_labels, fontsize=8)
            ax.yaxis.set_major_locator(ticker.MultipleLocator(1))
          plt.title(title)
          plt.savefig(fname)
          plt.close()
    elif args.output_format == "hdf":
      assert len(layers) == 1
      att_weights = kwargs["rec_%s" % layers[0]]
      hdf_writer.insert_batch(inputs=att_weights, seq_len={0: output_len, 1: encoder_len}, seq_tag=seq_tag)
    else:
      raise Exception("output format %r" % args.output_format)

  runner = Runner(engine=rnn.engine, dataset=dataset, batches=dataset_batch,
                  train=False, train_flag=bool(args.dropout) or args.train_flag,
                  extra_fetches=extra_fetches,
                  extra_fetches_callback=fetch_callback)
  runner.run(report_prefix="att-weights epoch %i" % rnn.engine.epoch)
  for layer in layers:
    stats[layer].dump(stream_prefix="Layer %r " % layer)
  if not runner.finalized:
    print("Some error occured, not finalized.")
    sys.exit(1)

  if hdf_writer:
    hdf_writer.close()
  rnn.finalize()


if __name__ == '__main__':
  main(sys.argv)



"""
Provides needed Git utilities to manage the local checked out packages.

The packages are stored in: ``~/returnn/pkg``
(Can be configured via env ``RETURNN_PKG_PATH``)

We might have multiple checkouts of the same repo.
In this case, we want to share the object files.
We are using `git worktree <https://git-scm.com/docs/git-worktree>`__ for this.

References:

  https://github.com/golang/go/blob/724d0720/src/cmd/go/internal/modfetch/codehost/git.go
  https://golang.org/ref/mod

Note about our terminology (esp vs Go terminology):

Go differentiates between packages, modules and repositories,
and even major semantic version sub-directories inside the modules.
We do not make this distinction.
You have a repository, and some specific commit,
and the file path inside the repo.
The file path might be sth like ``transformer-v3.py``
to note that it is incompatible from an earlier version.

Note about our ``version`` (revision) format:

During development, you would want to work in a live working tree.
There is a working tree in ``...repo``, and ``version=None`` refers to that.

Once development of one piece/module/package is ready,
you always should specify the version explicitly.
The format is ``<date>-<commit>``, where ``date`` should be like ``YYYYMMDDHHMMSS``
(can potentially be shorter up to ``YYYYMMDD``)
and ``commit`` is a revision, with minimum 7 numbers of digits.

Note about our ``repo`` format:

Currently always assumed to be remote, and then cached in our repo cache.
"""

import os
import typing
from subprocess import check_call, check_output, SubprocessError
import re
from . import common

_MinNumHashDigits = 7
_DefaultNumHashDigits = 12
_FullNumHashDigits = 40
_RevDigitsRe = re.compile("^[0123456789abcdef]*$")
_DateFormat = "%Y%m%d%H%M%S"
_DateFormatLen = 14
_DateFormatAllowedLens = (8, 10, 12, 14)  # Day only, or partial time.
# Not too complicated but also not totally dumb.
_DateRe = re.compile(r"^(\d\d\d\d)([01]\d)([0123]\d)(([012]\d)(([0-5]\d)([0-5]\d)?)?)?$")


def stat_repo(repo, version):
  """
  :param str repo: e.g. "github.com/rwth-i6/returnn-experiments"
  :param str|None version: e.g. "20211231-0123abcd0123"
  """
  repo_ = _get_repo(repo)
  repo_.get_work_dir(version)


def get_repo_path(repo, version, _report_dev_version_usage_stack_frame_depth=1):
  """
  :param str repo: e.g. "github.com/rwth-i6/returnn-experiments"
  :param str|None version: e.g. "20211231-0123abcd0123"
  :param int _report_dev_version_usage_stack_frame_depth:
  :return: path to repo
  :rtype: str
  """
  repo_ = _get_repo(repo)
  work_dir = repo_.get_work_dir(version)
  repo_path = work_dir.get_path()
  if not version:
    _report_usage_dev_version(repo_path=repo_path, stack_frame_depth=_report_dev_version_usage_stack_frame_depth + 1)
  return repo_path


def _simple_validate_repo_name(repo):
  """
  :param str repo:
  """
  assert ".." not in repo and ":" not in repo


def _main_repo_path(repo):
  """
  :param str repo:
  :return: main repo dir (which includes the Git objects)
  :rtype: str
  """
  return "%s/%s" % (common.package_path(), repo)


def _dev_repo_path(repo):
  """
  :param str repo:
  :return: dev working tree of a repo. currently the same as the main repo path
  :rtype: str
  """
  return _main_repo_path(repo)


def _repo_path(repo, version):
  """
  :param str repo:
  :param str|None version:
  :rtype: str
  """
  if not version:
    return _dev_repo_path(repo)
  return "%s@v%s" % (_main_repo_path(repo), version)


def _repo_remote_url(repo):
  """
  :param str repo:
  :rtype: str
  """
  # Note: See also Go documentation how it does this: https://golang.org/ref/mod#vcs-find
  # It uses https by default.
  # You can overwrite this via the global Git configuration.
  # Example for GitHub: put this into your ~/.gitconfig:
  # [url "git@github.com:"]
  #     insteadOf = https://github.com/
  p = repo.find("/")
  assert p >= 0
  host, path = repo[:p], repo[p:]
  return "https://%s:%s" % (host, path)


def _simple_validate_commit_rev(rev):
  """
  :param str rev:
  """
  assert len(rev) >= _MinNumHashDigits and _RevDigitsRe.match(rev)


def _simple_validate_date(date):
  """
  :param str date:
  """
  assert len(date) in _DateFormatAllowedLens and _DateRe.match(date)


def _simple_validate_version(version):
  """
  :param str|None version:
  """
  if not version:
    return
  date, rev = version.split("-")
  _simple_validate_date(date)
  _simple_validate_commit_rev(rev)


def _rev_from_version(version):
  """
  :param str version: e.g. "20211231-0123abcd0123"
  :return: e.g. "0123abcd0123"
  :rtype: str
  """
  p = version.rfind("-")
  if p < 0:  # ok, allow here
    _simple_validate_commit_rev(version)
    return version
  rev = version[p + 1:]
  _simple_validate_commit_rev(rev)
  return rev


def _version_from_date_and_rev(date, rev):
  """
  :param str date:
  :param str rev:
  """
  _simple_validate_commit_rev(rev)
  return "%s-%s" % (date, rev)


def _sys_git_clone_repo(repo):
  """
  :param str repo:
  """
  url = _repo_remote_url(repo)
  main_path = _main_repo_path(repo)
  common.logger.info("Git clone %s", repo)
  check_call(["git", "clone", url, main_path])


def _sys_git_fetch(repo):
  """
  :param str repo:
  """
  # Note: Just simply git fetch, no extra logic.
  # We might want to allow for named refs later...
  # We might want to extend this logic to be more like Go...
  # https://github.com/golang/go/blob/724d0720b3e/src/cmd/go/internal/modfetch/codehost/git.go#L366
  main_path = _main_repo_path(repo)
  common.logger.info("Git fetch %s", repo)
  check_call(["git", "fetch"], cwd=main_path)


def _sys_git_create_repo_workdir(repo, version):
  """
  :param str repo:
  :param str version:
  """
  main_path = _main_repo_path(repo)
  rev = _rev_from_version(version)
  version_path = _repo_path(repo, version)
  common.logger.info("Git add worktree %s: %s", repo, version)
  check_call(
    ["git", "worktree", "add", version_path, rev],
    cwd=main_path)


def _sys_git_stat_local_rev(repo, rev):
  """
  :param str repo:
  :param str rev:
  :return: (full_rev, date)
  :rtype: (str|None, str|None)
  """
  main_path = _main_repo_path(repo)
  try:
    out = check_output(
      ["git",
       "-c", "log.showsignature=false",
       "log", "-n1",
       "--format=format:%H %cd",
       "--date=format:%s" % _DateFormat,
       rev, "--"],
      cwd=main_path)
  except SubprocessError:
    return None, None
  out = out.decode("utf8")
  full_rev, date = out.split()
  assert full_rev.startswith(rev) and len(full_rev) == _FullNumHashDigits
  _simple_validate_commit_rev(full_rev)
  _simple_validate_date(date)
  return full_rev, date


class _Repo:
  def __init__(self, name):
    """
    :param str name: e.g. "github.com/rwth-i6/returnn-experiments"
    """
    _simple_validate_repo_name(name)
    self.name = name
    self._cloned = False
    self._loaded_local_work_dirs = False
    self._dev_work_dir = None  # type: typing.Optional[_RepoWorkDir]
    self._work_dirs = {}  # type: typing.Dict[str,_RepoWorkDir]  # commit-rev -> entry

  def get_dev_dir_path(self):
    """
    :rtype: str
    """
    self._clone()
    return _dev_repo_path(self.name)

  def _clone(self):
    if self._cloned:
      return
    main_path = _main_repo_path(self.name)
    if not os.path.exists(main_path):
      _sys_git_clone_repo(self.name)
      assert os.path.exists(main_path)
    self._cloned = True

  def _load_work_dirs(self):
    if self._loaded_local_work_dirs:
      return
    self._clone()
    main_path = _main_repo_path(self.name)
    p = main_path.rfind("/")
    assert p > 0
    dir_name, base_name = main_path[:p], main_path[p + 1:]
    prefix = base_name + "@v"
    for d in sorted(os.listdir(dir_name)):
      if not d.startswith(prefix):
        continue
      version = d[len(prefix):]
      rev = _rev_from_version(version)
      if rev in self._work_dirs:
        continue
      self._work_dirs[rev] = _RepoWorkDir(repo=self, version=version)
    self._loaded_local_work_dirs = True

  def _get_work_dir_stat_local(self, rev):
    """
    :param str rev:
    :rtype: _RepoWorkDir
    """
    full_rev, date = _sys_git_stat_local_rev(self.name, rev)
    if full_rev:
      default_rev = full_rev[:_DefaultNumHashDigits]
      assert default_rev not in self._work_dirs
      default_version = _version_from_date_and_rev(date=date, rev=default_rev)
      _sys_git_create_repo_workdir(self.name, version=default_version)
      work_dir = _RepoWorkDir(self, default_version)
      self._work_dirs[default_rev] = work_dir
      return work_dir
    return None

  def get_work_dir(self, version):
    """
    :param str|None version:
    :rtype: _RepoWorkDir
    :return: work dir. this makes sure that the work dir also exists
    """
    if not version:
      if self._dev_work_dir:
        return self._dev_work_dir
      self._clone()
      self._dev_work_dir = _RepoWorkDir(self, None)
      return self._dev_work_dir
    self._load_work_dirs()
    rev = _rev_from_version(version)
    if rev in self._work_dirs:
      work_dir = self._work_dirs[rev]
      work_dir.validate_version(version)
      return work_dir
    for rev_, work_dir in self._work_dirs.items():
      if rev_.startswith(rev) or (len(rev) > len(rev_) and rev.startswith(rev_)):
        work_dir.validate_version(version)
        self._work_dirs[rev] = work_dir
        return work_dir
    work_dir = self._get_work_dir_stat_local(rev)
    if work_dir:
      work_dir.validate_version(version)
      return work_dir
    _sys_git_fetch(self.name)
    work_dir = self._get_work_dir_stat_local(rev)
    assert work_dir, "Git repo %s, version %s unknown." % (self.name, version)
    work_dir.validate_version(version)
    return work_dir


class _RepoWorkDir:
  def __init__(self, repo, version):
    """
    :param str|_Repo repo:
    :param str|None version: (normalized)
    """
    if not isinstance(repo, _Repo):
      repo = _get_repo(repo)
    _simple_validate_version(version)
    self.repo = repo
    self.version = version
    if version:
      self.version_date, self.version_rev = version.split("-")
    else:
      self.version_date, self.version_rev = None, None

  def get_path(self):
    """
    :rtype: str
    """
    return _repo_path(self.repo.name, self.version)

  def validate_version(self, version):
    """
    :param str|None version:
    """
    if version:
      assert self.version
      date, rev = version.split("-")
      _simple_validate_date(date)
      _simple_validate_commit_rev(rev)
      if not self.version_date.startswith(date):
        raise common.InvalidVersion("Version %s, date does not match in %s" % (self.version, version))
      if len(rev) <= len(self.version_rev):
        if not self.version_rev.startswith(rev):
          raise common.InvalidVersion("Version %s, revision does not match in %s" % (self.version, version))
      else:
        if not rev.startswith(self.version_rev):
          raise common.InvalidVersion("Version %s, revision does not match in %s" % (self.version, version))
    else:
      if self.version:
        raise common.InvalidVersion("Development version, got %s" % (version,))


_repo_cache = {}  # type: typing.Dict[str,_Repo]


def _get_repo(repo):
  """
  :param str repo:
  :rtype: _Repo
  """
  assert isinstance(repo, str)
  obj = _repo_cache.get(repo)
  if obj:
    return obj
  obj = _Repo(repo)
  _repo_cache[repo] = obj
  return obj


def _report_usage_dev_version(repo_path, stack_frame_depth):
  """
  :param str repo_path:
  :param int stack_frame_depth:
  """
  common.logger.warn("Access to development working tree: %s", repo_path)
  from returnn.util.basic import try_get_stack_frame
  frame = try_get_stack_frame(depth=stack_frame_depth + 1)
  if frame:
    from returnn.util.better_exchook import get_source_code, add_indent_lines
    src = get_source_code(filename=frame.f_code.co_filename, lineno=frame.f_lineno, module_globals=frame.f_globals)
    common.logger.warn(
      "  Called from: %s:%i, code:\n%s",
      frame.f_code.co_filename, frame.f_lineno,
      add_indent_lines("    ", src.rstrip()))
  else:
    common.logger.warn("  (Could not get stack frame information from calling code.)")
  from returnn.util.basic import git_commit_date, git_commit_rev, git_is_dirty
  rev = git_commit_rev(git_dir=repo_path, length=_DefaultNumHashDigits)
  commit_date = git_commit_date(git_dir=repo_path)  # like "20190202.154527"
  commit_date = commit_date.replace(".", "")
  version = "%s-%s" % (commit_date[:8], rev[:_MinNumHashDigits])
  import_str = "v%s_%s" % (commit_date, rev)
  common.logger.warn("  Current version: %s (%s)", version, import_str)
  if git_is_dirty(git_dir=repo_path):
    common.logger.warn("  (Warning, code is dirty. Commit your recent changes.)")



import os
from theano.sandbox.cuda.basic_ops import (GpuContiguous, GpuFromHost)


def raw_variable(x):
  while x.owner is not None and (type(x.owner.op) == GpuContiguous or type(x.owner.op) == GpuFromHost):
    x = x.owner.inputs[0]
  return x


def get_c_support_code_common():
  base_path = os.path.dirname(__file__)
  with open(base_path + "/c_support_code_common.cpp") as f:
    return f.read()

def get_c_support_code_mdlstm():
  base_path = os.path.dirname(__file__)
  with open(base_path + "/c_support_code_mdlstm.cpp") as f:
    return f.read()

def get_c_support_code_cudnn():
  base_path = os.path.dirname(__file__)
  with open(base_path + "/c_support_code_cudnn.cpp") as f:
    return f.read()
